# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._context import Context

class KeyBlock:
    """Kerberos KeyBlock

    This class represents the contents of a key.

    Args:
        context: Krb5 context.
    """

    def __len__(self) -> int: ...
    @property
    def data(self) -> bytes:
        """The keyblock data."""
    @property
    def enctype(self) -> int:
        """The keyblock encryption type."""

def init_keyblock(
    context: Context,
    enctype: int,
    key: typing.Optional[bytes],
) -> KeyBlock:
    """Initialize a Key Block.

    Initalize a new keyblock and copy the key into the contents of that block.
    The key can be None or an empty byte string to represent the contents are
    not allocated.

    Args:
        context: Krb5 context.
        enctype: The encryption type of the keyblock.
        key: The data to place in the keyblock or None for an empty block.

    Returns:
        KeyBlock: The initialized keyblock.
    """



# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._exceptions import Krb5Error

from krb5._ccache cimport CCache
from krb5._context cimport Context
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    krb5_error_code krb5_cccol_cursor_free(
        krb5_context context,
        krb5_cccol_cursor *cursor
    ) nogil

    krb5_error_code krb5_cccol_cursor_new(
        krb5_context context,
        krb5_cccol_cursor *cursor,
    ) nogil

    krb5_error_code krb5_cccol_cursor_next(
        krb5_context context,
        krb5_cccol_cursor cursor,
        krb5_ccache *ccache,
    ) nogil

    krb5_error_code KRB5_CC_END


def cccol_iter(
    Context context not None,
) -> typing.Iterator[CCache]:
    cdef krb5_error_code err = 0
    cdef krb5_context ctx = context.raw
    cdef krb5_cccol_cursor cursor

    err = krb5_cccol_cursor_new(ctx, &cursor)
    if err:
        raise Krb5Error(context, err)

    try:
        while True:
            ccache = CCache(context)
            err = krb5_cccol_cursor_next(ctx, cursor, &ccache.raw)
            if err == KRB5_CC_END:  # Heimdal
                break
            elif err:
                raise Krb5Error(context, err)
            elif ccache.raw == NULL:  # MIT
                break

            yield ccache

    finally:
        err = krb5_cccol_cursor_free(ctx, &cursor)
        if err:
            raise Krb5Error(context, err)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import os
import typing

import k5test
import pytest

import krb5


# This could be extensive to do per function so just do it once and share that
@pytest.fixture(scope="session")
def realm() -> typing.Iterator[k5test.K5Realm]:
    test_realm = k5test.K5Realm()
    try:
        original_env: typing.Dict[str, typing.Optional[str]] = {}
        for k in test_realm.env.keys():
            original_env[k] = os.environ.pop(k, None)

        try:
            os.environ.update(test_realm.env)

            yield test_realm

        finally:
            for k, v in original_env.items():
                if v:
                    os.environ[k] = v
                else:
                    del os.environ[k]

    finally:
        test_realm.stop()
        del test_realm


@pytest.fixture(autouse=True)
def requires_api(request: typing.Any) -> None:
    marker = request.node.get_closest_marker("requires_api")
    if marker:
        api_name = marker.args[0]
        if not hasattr(krb5, api_name):
            pytest.skip(f"KRB5 API {api_name} not available on current environment")

    return


def pytest_configure(config: typing.Any) -> None:
    config.addinivalue_line(
        "markers",
        "requires_api(name): skip tests that don't have the required KRB5 API installed",
    )



# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from libc.stdlib cimport free

from krb5._exceptions import Krb5Error

from krb5._context cimport Context
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    # MIT and Heimdal have slightly different functions to init a keyblock
    """
    void pykrb5_keyblock_get(
        krb5_keyblock *key,
        krb5_enctype *enctype,
        size_t *length,
        char **data
    )
    {
    #if defined(HEIMDAL_XFREE)
        if (enctype != NULL) *enctype = key->keytype;
        if (length != NULL) *length = key->keyvalue.length;
        if (data != NULL) *data = key->keyvalue.data;
    #else
        if (enctype != NULL) *enctype = key->enctype;
        if (length != NULL) *length = key->length;
        if (data != NULL) *data = (char *)key->contents;
    #endif
    }

    krb5_error_code krb5_init_keyblock_generic(
        krb5_context context,
        krb5_enctype enctype,
        size_t length,
        const char *data,
        krb5_keyblock **out
    )
    {
    #if defined(HEIMDAL_XFREE)
        // While initialised here, krb5_free_keyblock will free this on deallocation
        krb5_keyblock *keyblock = NULL;
        keyblock = malloc(sizeof(krb5_keyblock));
        if (keyblock == NULL)
        {
            return ENOMEM;
        }

        *out = keyblock;
        return krb5_keyblock_init(context, enctype, data, length, keyblock);
    #else
        krb5_error_code err = 0;

        err = krb5_init_keyblock(context, enctype, length, out);
        if (err == 0 && length > 0)
        {
            memcpy((*out)->contents, data, length);
        }

        return err;
    #endif
    }
    """

    krb5_error_code krb5_init_keyblock_generic(
        krb5_context context,
        krb5_enctype enctype,
        size_t length,
        const char *data,
        krb5_keyblock **out,
    ) nogil

    krb5_error_code krb5_free_keyblock(
        krb5_context context,
        krb5_keyblock *val,
    ) nogil

    void pykrb5_keyblock_get(
        krb5_keyblock *key,
        krb5_enctype *enctype,
        size_t *length,
        char **data,
    ) nogil


cdef class KeyBlock:
    # cdef Context ctx
    # cdef krb5_keyblock *raw
    # cdef int needs_free

    def __cinit__(KeyBlock self, Context context, needs_free=1):
        self.ctx = context
        self.raw = NULL
        self.needs_free = needs_free

    def __dealloc__(KeyBlock self):
        if self.raw != NULL and self.needs_free:
            krb5_free_keyblock(self.ctx.raw, self.raw)
            self.raw = NULL

    def __len__(KeyBlock self) -> int:
        cdef size_t length
        pykrb5_keyblock_get(self.raw, NULL, &length, NULL)

        return length

    def __repr__(KeyBlock self) -> str:
        kwargs = [f"{k}={v}" for k, v in {
            'enctype': self.enctype,
            'length': len(self),
        }.items()]

        return f"KeyBlock({', '.join(kwargs)})"

    def __str__(KeyBlock self) -> str:
        return f"KeyBlock {self.enctype}"

    @property
    def data(KeyBlock self) -> bytes:
        cdef size_t length
        cdef char *data
        pykrb5_keyblock_get(self.raw, NULL, &length, &data)

        if length == 0:
            return b""
        else:
            return data[:length]

    @property
    def enctype(KeyBlock self) -> int:
        cdef krb5_enctype enctype
        pykrb5_keyblock_get(self.raw, &enctype, NULL, NULL)

        return enctype


def init_keyblock(
    Context context not None,
    krb5_enctype enctype,
    const unsigned char[:] key,
) -> KeyBlock:
    kb = KeyBlock(context)
    cdef krb5_error_code err = 0
    cdef size_t length = 0

    cdef const char *key_ptr = NULL
    if key is not None and len(key):
        length = len(key)
        key_ptr = <const char*>&key[0]

    err = krb5_init_keyblock_generic(context.raw, enctype, length, key_ptr, &kb.raw)
    if err:
        raise Krb5Error(context, err)

    return kb



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._context import Context
from krb5._creds_opt import GetInitCredsOpt

def get_init_creds_opt_set_default_flags(
    context: Context,
    opt: GetInitCredsOpt,
    appname: typing.Optional[bytes] = None,
    realm: typing.Optional[bytes] = None,
) -> None:
    """Set default configuration file flags.

    Set all the values in the options to the default values for the app and
    realm.

    Args:
        context: Krb5 context.
        ccache: The credential cache to set as the output.
        appname: The application name to get the default values for.
        realm: The realm as a fallback from appname.
    """



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from libc.stdint cimport int32_t, uint8_t, uint32_t


cdef extern from "python_krb5.h":
    """
    // The structures are different so cannot be explicitly defined in Cython. Use inline C to set the structs elements
    // by name.
    void pykrb5_set_krb5_data(
        krb5_data *data,
        size_t length,
        char *value
    )
    {
        data->length = length;
        data->data = value;
    }
    """

    ctypedef int32_t krb5_int32
    ctypedef krb5_int32 krb5_error_code
    ctypedef krb5_int32 krb5_deltat
    ctypedef krb5_int32 krb5_enctype
    ctypedef uint8_t krb5_octet
    ctypedef krb5_int32 krb5_timestamp
    ctypedef unsigned int krb5_boolean
    ctypedef unsigned int krb5_kvno

    ctypedef void *krb5_pointer;
    ctypedef krb5_pointer krb5_cc_cursor
    ctypedef krb5_pointer krb5_kt_cursor;

    cdef struct _krb5_context:
        pass
    ctypedef _krb5_context *krb5_context

    cdef struct _krb5_cccol_cursor:
        pass
    ctypedef _krb5_cccol_cursor *krb5_cccol_cursor

    cdef struct krb5_principal_data:
        pass
    ctypedef krb5_principal_data *krb5_principal
    ctypedef const krb5_principal_data *krb5_const_principal

    cdef struct _krb5_ccache:
        pass
    ctypedef _krb5_ccache *krb5_ccache

    cdef struct _krb5_get_init_creds_opt:
        pass
    ctypedef _krb5_get_init_creds_opt krb5_get_init_creds_opt

    cdef struct _krb5_init_creds_context:
        pass
    ctypedef _krb5_init_creds_context *krb5_init_creds_context

    cdef struct _krb5_keyblock:
        pass
    ctypedef _krb5_keyblock krb5_keyblock

    cdef struct _krb5_data:
        pass
    ctypedef _krb5_data krb5_data

    cdef struct _krb5_kt:
        pass
    ctypedef _krb5_kt *krb5_keytab

    cdef struct krb5_keytab_entry_st:
        pass
    ctypedef krb5_keytab_entry_st krb5_keytab_entry

    cdef struct _krb5_creds:
        pass
    ctypedef _krb5_creds krb5_creds

    cdef struct _krb5_prompt:
        char *prompt
        int hidden
        krb5_data *reply
    ctypedef _krb5_prompt krb5_prompt

    ctypedef krb5_error_code(*krb5_prompter_fct)(
        krb5_context context,
        void *data,
        const char *name,
        const char *banner,
        int num_prompts,
        krb5_prompt prompts[],
    )

    void pykrb5_set_krb5_data(
        krb5_data *data,
        size_t length,
        char *value,
    ) nogil



# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import k5test
import pytest

import krb5


def test_enctype_to_string(realm: k5test.K5Realm) -> None:
    expected = "AES-256 CTS mode with 96-bit SHA-1 HMAC" if realm.provider == "mit" else "aes256-cts-hmac-sha1-96"
    ctx = krb5.init_context()
    name = krb5.enctype_to_string(ctx, 18)
    assert name == expected


def test_enctype_to_string_invalid(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()

    expected_msg = "Invalid argument" if realm.provider == "mit" else "encryption type \\d+ not supported"
    with pytest.raises(krb5.Krb5Error, match=expected_msg):
        krb5.enctype_to_string(ctx, 1024)


def test_string_to_enctype() -> None:
    ctx = krb5.init_context()
    enctype = krb5.string_to_enctype(ctx, "aes256-cts-hmac-sha1-96")
    assert enctype == 18


def test_string_to_enctype_invalid(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    expected_msg = "Invalid argument" if realm.provider == "mit" else "encryption type invalid not supported"
    with pytest.raises(krb5.Krb5Error, match=expected_msg):
        krb5.string_to_enctype(ctx, "invalid")


@pytest.mark.requires_api("enctype_to_name")
def test_string_to_name() -> None:
    name = krb5.enctype_to_name(18)
    assert name == "aes256-cts-hmac-sha1-96"


@pytest.mark.requires_api("enctype_to_name")
def test_string_to_name_shortest() -> None:
    name = krb5.enctype_to_name(18, shortest=True)
    assert name == "aes256-cts"



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import collections
import typing

from krb5._exceptions import Krb5Error

from krb5._context cimport Context
from krb5._creds_opt cimport GetInitCredsOpt
from krb5._krb5_types cimport *
from krb5._kt cimport KeyTab
from krb5._principal cimport Principal


cdef extern from "python_krb5.h":
    void krb5_free_cred_contents(
        krb5_context context,
        krb5_creds *val,
    ) nogil

    krb5_error_code krb5_init_creds_get(
        krb5_context context,
        krb5_init_creds_context ctx,
    ) nogil

    krb5_error_code krb5_init_creds_get_creds(
        krb5_context context,
        krb5_init_creds_context ctx,
        krb5_creds *creds,
    ) nogil

    krb5_error_code krb5_init_creds_init(
        krb5_context context,
        krb5_principal client,
        krb5_prompter_fct prompert,
        void *data,
        krb5_deltat start_time,
        krb5_get_init_creds_opt *options,
        krb5_init_creds_context *ctx,
    ) nogil

    krb5_error_code krb5_get_init_creds_keytab(
        krb5_context context,
        krb5_creds *creds,
        krb5_principal client,
        krb5_keytab arg_keytab,
        krb5_deltat start_time,
        const char *in_tkt_service,
        krb5_get_init_creds_opt *k5_gic_options,
    ) nogil

    void krb5_init_creds_free(
        krb5_context context,
        krb5_init_creds_context ctx,
    ) nogil

    krb5_error_code krb5_get_init_creds_password(
        krb5_context context,
        krb5_creds *creds,
        krb5_principal client,
        const char *password,
        krb5_prompter_fct prompter,
        void *data,
        krb5_deltat start_time,
        const char *in_tkt_service,
        krb5_get_init_creds_opt *k5_gic_options,
    ) nogil

    krb5_error_code krb5_init_creds_set_keytab(
        krb5_context context,
        krb5_init_creds_context ctx,
        krb5_keytab keytab,
    ) nogil

    krb5_error_code krb5_init_creds_set_password(
        krb5_context context,
        krb5_init_creds_context ctx,
        const char *password,
    ) nogil


cdef class Creds:
    # cdef Context ctx
    # cdef krb5_creds raw
    # cdef int needs_free

    def __cinit__(Creds self, Context context):
        self.ctx = context
        self.needs_free = 0

    def __dealloc__(Creds self):
        if self.needs_free:
            krb5_free_cred_contents(self.ctx.raw, &self.raw)
            self.needs_free = 0

    def __str__(Creds self) -> str:
        return "Creds"


cdef class InitCredsContext:
    # cdef Context ctx
    # cdef krb5_init_creds_context raw

    def __cinit__(InitCredsContext self, Context context):
        self.ctx = context
        self.raw = NULL

    def __dealloc__(InitCredsContext self):
        if self.raw:
            krb5_init_creds_free(self.ctx.raw, self.raw)
            self.raw = NULL

    def __str__(InitCredsContext self) -> str:
        return "InitCredsContext"


cdef class Krb5Prompt:
    def init(
        self,
        name: typing.Optional[bytes],
        banner: typing.Optional[bytes],
        num_prompts: int,
    ) -> None:
        pass

    def prompt(
        self,
        msg: bytes,
        hidden: bool,
    ) -> bytes:
        raise NotImplementedError()


cdef krb5_error_code prompt_callback(
    krb5_context context,
    void *data,
    const char *name,
    const char *banner,
    int num_prompts,
    krb5_prompt *prompts,
) with gil:
    try:
        prompter = <Krb5Prompt>data

        py_name = None if name == NULL else <bytes>name
        py_banner = None if banner == NULL else <bytes>banner
        prompter.init(py_name, py_banner, num_prompts)

        replies = []
        for prompt in prompts[:num_prompts]:
            msg = <bytes>prompt.prompt
            hidden = prompt.hidden != 0

            reply = prompter.prompt(msg, hidden)
            if not isinstance(reply, bytes):
                return 1

            replies.append(reply)

        for idx, reply in enumerate(replies):
            pykrb5_set_krb5_data(prompts[idx].reply, len(reply), <char *>reply)

        return 0

    except Exception:
        return 1


def get_init_creds_keytab(
    Context context not None,
    Principal client not None,
    GetInitCredsOpt k5_gic_options not None,
    KeyTab keytab not None,
    int start_time = 0,
    const unsigned char[:] in_tkt_service = None,
) -> Creds:
    creds = Creds(context)
    cdef krb5_error_code err = 0

    cdef const char *in_tkt_service_ptr = NULL
    if in_tkt_service is not None and len(in_tkt_service):
        in_tkt_service_ptr = <const char*>&in_tkt_service[0]

    with nogil:
        err = krb5_get_init_creds_keytab(
            context.raw,
            &creds.raw,
            client.raw,
            keytab.raw,
            start_time,
            in_tkt_service_ptr,
            k5_gic_options.raw,
        )

    if err:
        raise Krb5Error(context, err)

    creds.needs_free = 1

    return creds


def get_init_creds_password(
    Context context not None,
    Principal client not None,
    GetInitCredsOpt k5_gic_options not None,
    const unsigned char[:] password = None,
    int start_time = 0,
    const unsigned char[:] in_tkt_service = None,
    prompter: typing.Optional[Krb5Prompt] = None,
) -> Creds:
    creds = Creds(context)
    cdef krb5_error_code err = 0

    cdef const char *password_ptr = NULL
    if password is not None and len(password):
        password_ptr = <const char*>&password[0]

    cdef krb5_prompter_fct callback = NULL
    cdef void *prompt_data = NULL
    if prompter is not None:
        callback = prompt_callback
        prompt_data = <void*>prompter

    cdef const char *in_tkt_service_ptr = NULL
    if in_tkt_service is not None and len(in_tkt_service):
        in_tkt_service_ptr = <const char*>&in_tkt_service[0]

    with nogil:
        err = krb5_get_init_creds_password(
            context.raw,
            &creds.raw,
            client.raw,
            password_ptr,
            callback,
            prompt_data,
            start_time,
            in_tkt_service_ptr,
            k5_gic_options.raw,
        )

    if err:
        raise Krb5Error(context, err)

    creds.needs_free = 1

    return creds


def init_creds_get(
    Context context not None,
    InitCredsContext ctx not None,
) -> None:
    cdef krb5_error_code err = 0

    with nogil:
        err = krb5_init_creds_get(context.raw, ctx.raw)

    if err:
        raise Krb5Error(context, err)


def init_creds_get_creds(
    Context context not None,
    InitCredsContext ctx not None,
) -> Creds:
    creds = Creds(context)
    cdef krb5_error_code err = 0

    err = krb5_init_creds_get_creds(context.raw, ctx.raw, &creds.raw)
    if err:
        raise Krb5Error(context, err)

    creds.needs_free = 1
    return creds


def init_creds_init(
    Context context not None,
    Principal client not None,
    GetInitCredsOpt k5_gic_options = None,
    int start_time = 0,
    prompter: typing.Optional[Krb5Prompt] = None,
) -> InitCredsContext:
    creds_ctx = InitCredsContext(context)
    cdef krb5_error_code err = 0

    cdef krb5_get_init_creds_opt *options = NULL
    if k5_gic_options:
        options = k5_gic_options.raw

    cdef krb5_prompter_fct callback = NULL
    cdef void *prompt_data = NULL
    if prompter is not None:
        callback = prompt_callback
        prompt_data = <void*>prompter

    with nogil:
        err = krb5_init_creds_init(
            context.raw,
            client.raw,
            callback,
            prompt_data,
            start_time,
            options,
            &creds_ctx.raw
        )

    if err:
        raise Krb5Error(context, err)

    return creds_ctx


def init_creds_set_keytab(
    Context context not None,
    InitCredsContext ctx not None,
    KeyTab keytab not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_init_creds_set_keytab(context.raw, ctx.raw, keytab.raw)
    if err:
        raise Krb5Error(context, err)


def init_creds_set_password(
    Context context not None,
    InitCredsContext ctx not None,
    const unsigned char[:] password,
) -> None:
    cdef krb5_error_code err = 0

    cdef const char *password_ptr = NULL
    if password is not None and len(password):
        password_ptr = <const char*>&password[0]
    else:
        raise ValueError("password must be set")

    err = krb5_init_creds_set_password(context.raw, ctx.raw, password_ptr)
    if err:
        raise Krb5Error(context, err)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from libc.stdint cimport uint32_t, uintptr_t
from libc.stdlib cimport free, malloc, realloc
from libc.string cimport strlen

from krb5._exceptions import Krb5Error
from krb5._principal import PrincipalParseFlags

from krb5._context cimport Context
from krb5._keyblock cimport KeyBlock
from krb5._krb5_types cimport *
from krb5._principal cimport Principal


cdef extern from "python_krb5.h":
    # Heimdal and MIT differ in their implementations
    """
    // The funtions do the structure manipulations in the C code as it can reference the struct entries by name rather
    // than an explicitly defined offset in Cython which differ cross implementations.

    void pykrb5_keytab_entry_get(
        krb5_keytab_entry *entry,
        krb5_principal *principal,
        krb5_timestamp *timestamp,
        krb5_kvno *vno,
        krb5_keyblock **key
    )
    {
        if (principal != NULL) *principal = entry->principal;
        if (timestamp != NULL) *timestamp = entry->timestamp;
        if (vno != NULL) *vno = entry->vno;
    #if defined(HEIMDAL_XFREE)
        if (key != NULL) *key = &entry->keyblock;
    #else
        if (key != NULL) *key = &entry->key;
    #endif
    }

    krb5_error_code krb5_kt_add_entry_generic(
        krb5_context context,
        krb5_keytab keytab,
        krb5_principal principal,
        krb5_kvno kvno,
        uint32_t timestamp,
        krb5_keyblock *key
    )
    {
        krb5_keytab_entry entry;

    #if defined(HEIMDAL_XFREE)
        entry.principal = principal;
        entry.vno = kvno;
        entry.keyblock.keytype = key->keytype;
        entry.keyblock.keyvalue = key->keyvalue;
        entry.timestamp = timestamp;
    #else
        entry.magic = 0;
        entry.principal = principal;
        entry.timestamp = timestamp;
        entry.vno = kvno;
        entry.key.magic = key->magic;
        entry.key.enctype = key->enctype;
        entry.key.length = key->length;
        entry.key.contents = key->contents;
    #endif

        return krb5_kt_add_entry(context, keytab, &entry);
    }

    krb5_error_code krb5_kt_get_type_generic(
        krb5_context context,
        krb5_keytab keytab,
        char **prefix,
        size_t prefixsize
    )
    {
    // MIT won't have this defined and our header sets it to -1 if it wasn't defined
    #if KRB5_KT_PREFIX_MAX_LEN == -1
        *prefix = (char *)krb5_kt_get_type(context, keytab);
        return 0;
    #else
        return krb5_kt_get_type(context, keytab, *prefix, prefixsize);
    #endif
    }

    krb5_error_code krb5_kt_free_entry_generic(
        krb5_context context,
        krb5_keytab_entry *entry
    )
    {
    // MIT has deprecated krb5_kt_free_entry in favour of this function
    #if defined(HEIMDAL_XFREE)
        return krb5_kt_free_entry(context, entry);
    #else
        return krb5_free_keytab_entry_contents(context, entry);
    #endif
    }
    """

    krb5_error_code krb5_kt_add_entry_generic(
        krb5_context context,
        krb5_keytab keytab,
        krb5_principal principal,
        krb5_kvno kvno,
        uint32_t timestamp,
        krb5_keyblock *key,
    ) nogil

    krb5_error_code krb5_kt_close(
        krb5_context context,
        krb5_keytab keytab,
    ) nogil

    krb5_error_code krb5_kt_default(
        krb5_context context,
        krb5_keytab *id,
    ) nogil

    krb5_error_code krb5_kt_default_name(
        krb5_context context,
        char *name,
        int name_size,
    ) nogil

    krb5_error_code krb5_kt_end_seq_get(
        krb5_context context,
        krb5_keytab keytab,
        krb5_kt_cursor *cursor,
    ) nogil

    krb5_error_code krb5_kt_free_entry_generic(
        krb5_context context,
        krb5_keytab_entry *entry,
    ) nogil

    krb5_error_code krb5_kt_get_entry(
        krb5_context context,
        krb5_keytab keytab,
        krb5_const_principal principal,
        krb5_kvno vno,
        krb5_enctype enctype,
        krb5_keytab_entry *entry,
    ) nogil

    krb5_error_code krb5_kt_get_name(
        krb5_context context,
        krb5_keytab keytab,
        char *name,
        unsigned int namelen,
    ) nogil

    # See inline C code
    krb5_error_code krb5_kt_get_type_generic(
        krb5_context context,
        krb5_keytab keytab,
        char **prefix,
        size_t prefixsize,
    ) nogil

    krb5_error_code krb5_kt_next_entry(
        krb5_context context,
        krb5_keytab keytab,
        krb5_keytab_entry *entry,
        krb5_kt_cursor *cursor,
    ) nogil

    krb5_error_code krb5_kt_read_service_key(
        krb5_context context,
        krb5_pointer keyprocarg,
        krb5_principal principal,
        krb5_kvno vno,
        krb5_enctype enctype,
        krb5_keyblock **key,
    ) nogil

    krb5_error_code krb5_kt_remove_entry(
        krb5_context context,
        krb5_keytab id,
        krb5_keytab_entry *entry
    ) nogil

    krb5_error_code krb5_kt_resolve(
        krb5_context context,
        const char *name,
        krb5_keytab *ktid,
    ) nogil

    krb5_error_code krb5_kt_start_seq_get(
        krb5_context context,
        krb5_keytab keytab,
        krb5_kt_cursor *cursor,
    ) nogil

    void pykrb5_keytab_entry_get(
        krb5_keytab_entry *entry,
        krb5_principal *principal,
        krb5_timestamp *timestamp,
        krb5_kvno *vno,
        krb5_keyblock **key,
    ) nogil

    krb5_error_code KRB5_CONFIG_NOTENUFSPACE
    krb5_error_code KRB5_KT_END
    krb5_error_code KRB5_KT_NAME_TOOLONG
    krb5_error_code KRB5_KT_PREFIX_MAX_LEN


cdef class KeyTab:
    # cdef Context ctx
    # cdef krb5_keytab raw

    def __cinit__(KeyTab self, Context context):
        self.ctx = context
        self.raw = NULL

    def __dealloc__(KeyTab self):
        if self.raw:
            krb5_kt_close(self.ctx.raw, self.raw)
            self.raw = NULL

    def __iter__(KeyTab self) -> typing.Iterator["KeyTabEntry"]:
        cdef krb5_error_code err = 0
        cdef krb5_kt_cursor cursor

        err = krb5_kt_start_seq_get(self.ctx.raw, self.raw, &cursor)
        if err:
            raise Krb5Error(self.ctx, err)

        try:
            while True:
                entry = KeyTabEntry(self.ctx)
                err = krb5_kt_next_entry(self.ctx.raw, self.raw, &entry.raw, &cursor)
                if err == KRB5_KT_END:
                    break
                elif err:
                    raise Krb5Error(self.ctx, err)

                entry.needs_free = 1
                yield entry

        finally:
            err = krb5_kt_end_seq_get(self.ctx.raw, self.raw, &cursor)
            if err:
                raise Krb5Error(self.ctx, err)

    @property
    def addr(self) -> typing.Optional[int]:
        if self.raw:
            return <uintptr_t>self.raw

    @property
    def name(KeyTab self) -> typing.Optional[bytes]:
        if self.raw:
            # MIT returns the type + name, this strips out the type to ensure
            # consistent behaviour across providers.
            name = kt_get_name(self.ctx, self)
            kt_type = self.kt_type

            if name.startswith(kt_type + b":"):
                name = name[len(kt_type) + 1:]

            return name

    @property
    def kt_type(KeyTab self) -> typing.Optional[bytes]:
        if self.raw:
            return kt_get_type(self.ctx, self)

    def __repr__(KeyTab self) -> str:
        if self.raw:
            kwargs = [f"{k}={v}" for k, v in {
                'kt_type': self.kt_type.decode('utf-8'),
                'name': self.name.decode('utf-8'),
            }.items()]

            return f"KeyTab({', '.join(kwargs)})"

        else:
            return "KeyTab(NULL)"

    def __str__(KeyTab self) -> str:
        if self.raw:
            return f"{self.kt_type.decode('utf-8')}:{self.name.decode('utf-8')}"

        else:
            return "NULL"


cdef class KeyTabEntry:
    # cdef Context ctx
    # cdef krb5_keytab_entry raw
    # cdef int needs_free

    def __cinit__(KeyTabEntry self, Context context):
        self.ctx = context
        self.needs_free = 0

    def __dealloc__(KeyTabEntry self):
        if self.needs_free:
            krb5_kt_free_entry_generic(self.ctx.raw, &self.raw)
            self.needs_free = 0

    def __repr__(KeyTabEntry self) -> str:
        kwargs = [f"{k}={v}" for k, v in {
            'principal': repr(self.principal),
            'timestamp': self.timestamp,
            'kvno': self.kvno,
            'key': repr(self.key),
        }.items()]

        return f"KeyTabEntry({', '.join(kwargs)})"

    def __str__(KeyTabEntry self) -> str:
        return f"KVNO {self.kvno} {self.principal!s}"

    @property
    def key(KeyTabEntry self) -> KeyBlock:
        kb = KeyBlock(self.ctx, needs_free=0)
        pykrb5_keytab_entry_get(&self.raw, NULL, NULL, NULL, &kb.raw)

        return kb

    @property
    def kvno(KeyTabEntry self) -> int:
        cdef krb5_kvno kvno
        pykrb5_keytab_entry_get(&self.raw, NULL, NULL, &kvno, NULL)

        return kvno

    @property
    def principal(KeyTabEntry self) -> Principal:
        principal = Principal(self.ctx, PrincipalParseFlags.none, needs_free=0)
        pykrb5_keytab_entry_get(&self.raw, &principal.raw, NULL, NULL, NULL)

        return principal

    @property
    def timestamp(KeyTabEntry self) -> int:
        cdef krb5_timestamp timestamp
        pykrb5_keytab_entry_get(&self.raw, NULL, &timestamp, NULL, NULL)

        return timestamp


def kt_add_entry(
    Context context not None,
    KeyTab keytab not None,
    Principal principal not None,
    krb5_kvno kvno,
    uint32_t timestamp,
    KeyBlock keyblock not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_kt_add_entry_generic(context.raw, keytab.raw, principal.raw, kvno, timestamp, keyblock.raw)
    if err:
        raise Krb5Error(context, err)


def kt_default(
    Context context not None,
) -> KeyTab:
    kt = KeyTab(context)
    cdef krb5_error_code err = 0

    err = krb5_kt_default(context.raw, &kt.raw)
    if err:
        raise Krb5Error(context, err)

    return kt


def kt_default_name(
    Context context not None,
) -> bytes:
    cdef krb5_error_code err = KRB5_CONFIG_NOTENUFSPACE
    buffer_size = 8192
    buffer_length = buffer_size
    cdef char *buffer = <char *>malloc(buffer_length)
    if not buffer:
        raise MemoryError()

    try:
        while err == KRB5_CONFIG_NOTENUFSPACE:
            err = krb5_kt_default_name(context.raw, buffer, buffer_length)

            if err == KRB5_CONFIG_NOTENUFSPACE:
                buffer_length += buffer_size

                # Use a temp var to ensure buffer is always something valid to be freed
                new_buffer = <char *>realloc(buffer, buffer_length)
                if not new_buffer:
                    raise MemoryError()
                buffer = new_buffer

            elif err:
                raise Krb5Error(context, err)

            else:
                name_len = strlen(buffer)
                return <bytes>buffer[:name_len]

    finally:
        free(buffer)


def kt_get_entry(
    Context context not None,
    KeyTab keytab not None,
    Principal principal not None,
    krb5_kvno kvno=0,
    krb5_enctype enctype=0,
) -> KeyTabEntry:
    cdef KeyTabEntry entry = KeyTabEntry(context)
    cdef krb5_error_code err = 0

    err = krb5_kt_get_entry(context.raw, keytab.raw, principal.raw, kvno, enctype, &entry.raw)
    if err:
        raise Krb5Error(context, err)

    return entry


def kt_get_name(
    Context context not None,
    KeyTab keytab not None,
) -> bytes:
    cdef krb5_error_code err = KRB5_KT_NAME_TOOLONG
    buffer_size = 8192
    buffer_length = buffer_size
    cdef char *buffer = <char *>malloc(buffer_length)
    if not buffer:
        raise MemoryError()

    try:
        while err == KRB5_KT_NAME_TOOLONG:
            err = krb5_kt_get_name(context.raw, keytab.raw, buffer, buffer_length)

            # Heimdal does not define KRB5_KT_NAME_TOOLONG so fail on the first try.
            if err and (err != KRB5_KT_NAME_TOOLONG or KRB5_KT_NAME_TOOLONG == 1):
                raise Krb5Error(context, err)

            elif err == KRB5_KT_NAME_TOOLONG:
                buffer_length += buffer_size

                # Use a temp var to ensure buffer is always something valid to be freed
                new_buffer = <char *>realloc(buffer, buffer_length)
                if not new_buffer:
                    raise MemoryError()
                buffer = new_buffer

            else:
                name_len = strlen(buffer)
                return <bytes>buffer[:name_len]

    finally:
        free(buffer)


def kt_get_type(
    Context context not None,
    KeyTab keytab not None,
) -> bytes:
    cdef krb5_error_code err = 0
    cdef char *buffer = NULL

    if KRB5_KT_PREFIX_MAX_LEN == -1:
        # MIT Kerberos just returns a const char* which shouldn't be freed
        krb5_kt_get_type_generic(context.raw, keytab.raw, &buffer, 0)
        return <bytes>buffer

    else:
        # Heimdal requires the caller to allocate and free the memory. Use the defined max len for the prefix.
        buffer_length = KRB5_KT_PREFIX_MAX_LEN
        buffer = <char *>malloc(buffer_length)
        if not buffer:
            raise MemoryError()

        try:
            err = krb5_kt_get_type_generic(context.raw, keytab.raw, &buffer, buffer_length)
            if err:
                raise Krb5Error(context, err)

            type_len = strlen(buffer)
            return <bytes>buffer[:type_len]

        finally:
            free(buffer)


def kt_read_service_key(
    Context context not None,
    const unsigned char[:] name,
    Principal principal not None,
    krb5_kvno kvno = 0,
    krb5_enctype enctype = 0,
) -> KeyBlock:
    kb = KeyBlock(context)

    cdef krb5_error_code err = 0
    cdef krb5_pointer name_ptr = NULL
    if len(name):
        name_ptr = <krb5_pointer>&name[0]
    else:
        raise ValueError("KeyTab must be set")

    err = krb5_kt_read_service_key(context.raw, name_ptr, principal.raw, kvno, enctype, &kb.raw)
    if err:
        raise Krb5Error(context, err)

    return kb


def kt_remove_entry(
    Context context not None,
    KeyTab keytab not None,
    KeyTabEntry entry not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_kt_remove_entry(context.raw, keytab.raw, &entry.raw)
    if err:
        raise Krb5Error(context, err)


def kt_resolve(
    Context context not None,
    const unsigned char[:] name not None,
) -> KeyTab:
    kt = KeyTab(context)
    cdef krb5_error_code err = 0

    cdef const char *name_ptr = NULL
    if len(name):
        name_ptr = <const char*>&name[0]
    else:
        raise ValueError("KeyTab must be set")

    err = krb5_kt_resolve(context.raw, name_ptr, &kt.raw)
    if err:
        raise Krb5Error(context, err)

    return kt



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from libc.stdint cimport int32_t

from krb5._context cimport Context
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    const char *krb5_get_error_message(
        krb5_context ctx,
        krb5_error_code code,
    ) nogil

    void krb5_free_error_message(
        krb5_context ctx,
        const char *msg,
    ) nogil

    krb5_error_code KRB5_KT_NAME_TOOLONG
    # krb5_error_code KRB5_CONFIG_NOTENUFSPACE


cdef str get_error_message(
    krb5_context ctx,
    krb5_error_code err,
):
    cdef const char *err_msg = NULL

    err_msg = krb5_get_error_message(ctx, err)
    try:
        return err_msg.decode('utf-8')

    finally:
        krb5_free_error_message(ctx, err_msg)


class Krb5Error(Exception):

    def __init__(
        self,
        context: Context,
        err_code: int,
    ) -> None:
        self.err_code = err_code

        msg = get_error_message(context.raw, err_code)
        super().__init__(f"{msg} {self.err_code}")



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import os

import k5test
import pytest

import krb5


def test_init_context() -> None:
    context = krb5.init_context()
    assert context is not None
    assert isinstance(context, krb5.Context)
    assert str(context) == "Krb5Context"


def test_set_default_realm(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()

    default = krb5.get_default_realm(ctx)
    assert default == realm.realm.encode()

    krb5.set_default_realm(ctx, b"NEW.REALM")
    default = krb5.get_default_realm(ctx)
    assert default == b"NEW.REALM"

    krb5.set_default_realm(ctx, None)
    default = krb5.get_default_realm(ctx)
    assert default == realm.realm.encode()


@pytest.mark.requires_api("init_secure_context")
def test_init_secure_context() -> None:
    context = krb5.init_secure_context()
    assert context is not None
    assert isinstance(context, krb5.Context)
    assert str(context) == "Krb5Context"



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._context import Context
from krb5._creds_opt import GetInitCredsOpt
from krb5._kt import KeyTab
from krb5._principal import Principal

class Creds:
    """Kerberos Credentials object.

    This class represents Kerberos credentials.

    Args:
        context: Krb5 context.
    """

class InitCredsContext:
    """Kerberos Initial Credentials object.

    This class represents Kerberos context for acquiring initial credentials.

    Args:
        context: Krb5 context.
    """

class Krb5Prompt:
    """Base class used to prompt the user for input.

    This is the base class the caller should use to create a prompter class
    that is passed into :meth:`get_init_creds_password` and
    :meth:`init_creds_init`. It is used to handle prompts from the Kerberos
    API, such as changing a password or password input.

    At a minimum the `prompt` function should be implemented to return the
    bytes for the prompt.
    """

    def init(
        self,
        name: typing.Optional[bytes],
        banner: typing.Optional[bytes],
        num_prompts: int,
    ) -> None:
        """Initialise the prompt.

        Initialises the prompt by passing in the optional name, banner, and the
        number of prompts expected.

        Args:
            name: Optional name for the prompt session which should be
                displayed to the user.
            banner: Optional message that indicates what sort of input is
                expected from the user which should be displayed to the user.
            num_prompts: Number of times :meth:`prompt` is to be called for
                this prompt session.
        """
    def prompt(
        self,
        msg: bytes,
        hidden: bool,
    ) -> bytes:
        """Prompt user for a response.

        Prompt the user for a response.

        Args:
            msg: The message to display for this prompt.
            hidden: Set to ``True`` to indicate the response should not be
                printed back to the user.

        Returns:
            bytes: The response for the prompt.
        """

def get_init_creds_keytab(
    context: Context,
    client: Principal,
    k5_gic_options: GetInitCredsOpt,
    keytab: KeyTab,
    start_time: int = 0,
    in_tkt_service: typing.Optional[bytes] = None,
) -> Creds:
    """Get initial credentials using a key table.

    Requests the KDC for credential using a client key stored in the key table
    specified.

    Args:
        context: Krb5 context.
        client: The client principal the credentials are for.
        k5_gic_options: The initial credentials options.
        keytab: The keytab to use when getting the credential.
        start_time: Time when the ticket becomes valid, 0 for now.
        in_tkt_service: The service name of the initial credentials.

    Returns:
        Creds: The retrieved credentials.
    """

def get_init_creds_password(
    context: Context,
    client: Principal,
    k5_gic_options: GetInitCredsOpt,
    password: typing.Optional[bytes] = None,
    start_time: int = 0,
    in_tkt_service: typing.Optional[bytes] = None,
    prompter: typing.Optional[Krb5Prompt] = None,
) -> Creds:
    """Get initial credential using a password.

    Requests the KDC for credentials using a password.

    Args:
        context: Krb5 context.
        client: The client principal the credentials are for.
        k5_gic_options: The initial credentials options.
        password: The password to use - set to ``None`` to use the prompter.
        start_time: Time when the ticket becomes valid, 0 for now.
        in_tkt_service: The service name of the initial credentials.
        prompter: The prompter instance that will handle interactive prompts.

    Returns:
        Creds: The retrieved credentials.
    """

def init_creds_get(
    context: Context,
    ctx: InitCredsContext,
) -> None:
    """Acquire credentials using initial creds context.

    Obtains credentials using the initial creds context from
    :meth:`init_creds_init`. The credentials can be retrieved with
    :meth:`init_creds_get_creds`.

    Args:
        context: Krb5 context.
        ctx: Initial credentials context.
    """

def init_creds_get_creds(
    context: Context,
    ctx: InitCredsContext,
) -> Creds:
    """Retreived acquired creds from an initial creds context.

    Gets the acquired creds from a successful call of :meth:`init_creds_get`.

    Args:
        context: Krb5 context.
        ctx: Initial credentials context.

    Returns:
        Creds: The acquired credentials.
    """

def init_creds_init(
    context: Context,
    client: Principal,
    k5_gic_options: typing.Optional[GetInitCredsOpt] = None,
    start_time: int = 0,
    prompter: typing.Optional[Krb5Prompt] = None,
) -> InitCredsContext:
    """Get initial acquiring credential context.

    Creates a new context used for acquiring initial credentials.

    Args:
        context: Krb5 context.
        client: The client principal the credentials are for.
        k5_gic_options: The initial credentials options.
        start_time: Time when the ticket becomes valid, 0 for now.
        prompter: The prompter instance that will handle interactive prompts.

    Returns:
        InitCredsContext: The retrieved acquiring initial credentials context.
    """

def init_creds_set_keytab(
    context: Context,
    ctx: InitCredsContext,
    keytab: KeyTab,
) -> None:
    """Set the keytab used for acquiring initial credentials.

    Sets the keytab used to contruct the client key for an initial
    credentials request. See :meth:`get_init_creds_keytab` for a simpler
    way to get credentials with a keytab.

    Args:
        context: Krb5 context.
        ctx: Initial credentials context.
        keytab: The keytab to set.
    """

def init_creds_set_password(
    context: Context,
    ctx: InitCredsContext,
    password: bytes,
) -> None:
    """Set the password used for acquiring initial credentials.

    Sets the password used to contruct the client key for an initial
    credentials request. See :meth:`get_init_creds_password` for a simpler
    way to get credentials with a password.

    Args:
        context: Krb5 context.
        ctx: Initial credentials context.
        password: The password to set.
    """



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._context import Context
from krb5._creds import Creds
from krb5._principal import Principal

class CCache:
    """Kerberos CCache

    This class represents a Credential Cache object.

    Args:
        context: Krb5 context.
    """

    def __iter__(self) -> typing.Iterator[Creds]:
        """Iterate credentials in a ccache."""
    @property
    def addr(self) -> typing.Optional[int]:
        """The raw krb5_ccache pointer address of this credential cache."""
    @property
    def name(self) -> typing.Optional[bytes]:
        """The name/residual of the credential cache."""
    @property
    def principal(self) -> typing.Optional[Principal]:
        """Default client principal of the credential cache."""
    @property
    def cache_type(self) -> typing.Optional[bytes]:
        """The type of the credential cache."""

def cc_default(
    context: Context,
) -> CCache:
    """Resolve the default credential cache name.

    Create a handle to the default credential cache as given by
    :meth:`cc_default_name()`.

    Args:
        context: Krb5 context.

    Returns:
        CCache: The opened credential cache.
    """

def cc_default_name(
    context: Context,
) -> bytes:
    """Get the name of the default credential cache.

    The name of the default credential cache. This value consults a wide range
    of sources such as the ``KRB5CCNAME`` env var, config settings, build time
    variables, etc.

    Args:
        context: Krb5 context.

    Returns:
        bytes: The name of the default credential cache.
    """

def cc_destroy(
    context: Context,
    cache: CCache,
) -> None:
    """Destroy a credential cache.

    Destroys the contents of the cache and closes the handle to it.

    Args:
        context: Krb5 context.
        cache: The credential cache to destroy.
    """

def cc_get_name(
    context: Context,
    cache: CCache,
) -> bytes:
    """Retrieve the name of a credential cache.

    Gets the name of the credential, but not the type of a credential cache.

    Args:
        context: Krb5 context.
        cache: The credential cache to query.

    Returns:
        bytes: The credential cache name.
    """

def cc_get_principal(
    context: Context,
    cache: CCache,
) -> Principal:
    """Get the default principal of a credential cache.

    Gets the default principal of a credential as set by
    :meth:`cc_initialize()`.

    Args:
        context: Krb5 context.
        cache: The credential cache to query.

    Returns:
        Principal: The default principal of the cache.
    """

def cc_get_type(
    context: Context,
    cache: CCache,
) -> bytes:
    """Retrieve the type of a credential cache.

    Gets the credential cache type.

    Args:
        context: Krb5 context.
        cache: The credential cache to query.

    Returns:
        bytes: The credential cache type.
    """

def cc_initialize(
    context: Context,
    cache: CCache,
    principal: Principal,
) -> None:
    """Initialize a credential cache.

    Destroy any existing contents of a cache and initialize it for the default
    principal specified.

    Args:
        context: Krb5 context.
        cache: The cache to destroy and recreate.
        principal: The default principal of the cache.
    """

def cc_new_unique(
    context: Context,
    cred_type: bytes,
    hint: typing.Optional[bytes] = None,
) -> CCache:
    """Create a new unique credential cache.

    Create a new credential cache of the specified type with a unique name.

    Args:
        context: Krb5 context.
        cred_type: The credential cache type to create.
        hint: Unused.

    Returns:
        CCache: The created credential cache.
    """

def cc_resolve(
    context: Context,
    name: bytes,
) -> CCache:
    """Resolve a credential cache name.

    Resolves/opens a credential cache by the name specified. The name should be
    in the form ``type:residual`` where the type is known to the krb5 library
    being called. If the name does not contain a colon then it is interpreted
    as a file name (``FILE:*``).

    Args:
        context: Krb5 context.
        name: The name of the cache to resolve.

    Returns:
        CCache: The credential cache that was resolved.
    """

def cc_set_default_name(
    context: Context,
    name: typing.Optional[bytes],
) -> None:
    """Set the default ccache name.

    Set the default credential cache name to the name specified for future
    operations using the context. If name is `None` or an empty byte string
    this will clear any previous application-set default name and forget any
    cached value of the default name for the context.

    Args:
        context: Krb5 context.
        name: The default credential name or `None` to reset back to the config
            defaults.
    """

def cc_store_cred(
    context: Context,
    cache: CCache,
    creds: Creds,
) -> None:
    """Store credentials in a credential cache.

    Stores the credentials into the credential cache specified.

    Args:
        context: Krb5 context.
        cache: The credential cache to store the creds into.
        creds: The credentials to store.
    """

def cc_switch(
    context: Context,
    cache: CCache,
) -> None:
    """Switch primary cache in a collection.

    If the type of cache supports it, set the cache to be the primary
    credential cache for the collection it belongs to.

    Args:
        context: Krb5 context.
        cache: The credential cache to set as the primary in its collection.
    """



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._context cimport Context
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    krb5_error_code krb5_init_secure_context(
        krb5_context *context,
    ) nogil


def init_secure_context() -> Context:
    context = Context()
    krb5_init_secure_context(&context.raw)

    return context



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._context cimport Context
from krb5._krb5_types cimport *


cdef class CCache:
    cdef Context ctx
    cdef krb5_ccache raw



# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from libc.stdlib cimport free, malloc

from krb5._exceptions import Krb5Error

from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    krb5_error_code krb5_enctype_to_name(
        krb5_enctype enctype,
        krb5_boolean shortest,
        char *buffer,
        size_t buflen,
    ) nogil


def enctype_to_name(
    krb5_enctype enctype,
    krb5_boolean shortest = False,
) -> str:
    cdef krb5_error_code err = 0
    cdef void *buffer = malloc(100)
    if buffer == NULL:
        raise MemoryError()

    err = krb5_enctype_to_name(enctype, shortest, <char *>buffer, 100)
    if err:
        raise ValueError("Invalid encryption type")

    try:
        return (<char *>buffer).decode("utf-8")
    finally:
        free(buffer)



# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import krb5


def test_init_keyblock_empty() -> None:
    ctx = krb5.init_context()
    kb = krb5.init_keyblock(ctx, 0, None)

    assert len(kb) == 0
    assert kb.enctype == 0
    assert kb.data == b""
    assert str(kb) == "KeyBlock 0"
    assert repr(kb) == "KeyBlock(enctype=0, length=0)"


def test_init_keyblock_data() -> None:
    ctx = krb5.init_context()
    kb = krb5.init_keyblock(ctx, 17, b"\xff" * 16)

    assert len(kb) == 16
    assert kb.enctype == 17
    assert kb.data == b"\xff" * 16
    assert str(kb) == "KeyBlock 17"
    assert repr(kb) == "KeyBlock(enctype=17, length=16)"



#!/usr/bin/env python

# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

"""Builds the project.

Logic has mostly been derived from python-gssapi - thanks to them for all the
hard work.

https://github.com/pythongssapi/python-gssapi/blob/main/setup.py
"""

import ctypes
import ctypes.util
import os
import os.path
import platform
import shlex
import subprocess
import sys
import typing

from Cython.Build import cythonize
from setuptools import Extension, setup

SKIP_EXTENSIONS = os.environ.get("KRB5_SKIP_EXTENSIONS", "false").lower() == "true"
SKIP_MODULE_CHECK = os.environ.get("KRB5_SKIP_MODULE_CHECK", "false").lower() == "true"
CYTHON_LINETRACE = os.environ.get("KRB5_CYTHON_TRACING", "false").lower() == "true"


def run_command(*args: str) -> str:
    stdout = subprocess.check_output(args, shell=True)
    return stdout.decode("utf-8").strip()


def make_extension(
    name: str,
    module: ctypes.CDLL,
    canary: typing.Optional[str] = None,
    **kwargs: typing.Any,
) -> Extension:
    source = os.path.join("src", name.replace(".", os.sep)) + ".pyx"

    if not SKIP_MODULE_CHECK and canary and not hasattr(module, canary):
        print(f"Skipping {source} as it is not supported by the selected Kerberos implementation.")
        return

    if not os.path.exists(source):
        raise FileNotFoundError(source)

    print(f"Compiling {source}")
    return Extension(
        name=name,
        sources=[source],
        **kwargs,
    )


def get_krb5_config() -> str:
    """Gets the path to the krb5-config binary.

    Determines the path to the krb5-confing to use for detecting linker and
    compiler args on the current platform. The default behaviour is to use
    the value of the env var ``KRB5_KRB5CONFIG`` defaulting to just
    krb5-conf in the path.

    FreeBSD is slightly different as it favours ``/usr/local/bin/krb5-config``
    over ``/usr/bin/krb5-config`` due to the latter being based on an ancient
    Heimdal version.

    Returns:
        str: Path to krb5-conf.
    """
    kc_env = os.environ.get("KRB5_KRB5CONFIG", "")
    kc = kc_env if kc_env else "krb5-config"

    if sys.platform.startswith("freebsd") and not kc_env:
        # FreeBSD does $PATH backward, for our purposes.  That is, the package
        # manager's version of the software is in /usr/local, which is in PATH
        # *after* the version in /usr.  We prefer the package manager's version
        # because the Heimdal in base is truly ancient, but this can be overridden
        # - either in the "normal" fashion by putting something in PATH in front
        # of it, or by removing /usr/local from PATH.
        bins = []
        for entry in os.environ.get("PATH", "").split(os.pathsep):
            p = os.path.join(entry, "krb5-config")
            if not os.path.exists(p):
                continue
            bins.append(p)

        if len(bins) > 1 and bins[0] == "/usr/bin/krb5-config" and "/usr/local/bin/krb5-config" in bins:
            kc = "/usr/local/bin/krb5-config"

    return kc


def get_krb5_lib_path(
    libraries: typing.List[str],
    library_dirs: typing.List[str],
    link_args: typing.List[str],
    macos_native: bool,
) -> str:
    """Gets the path to the main libkrb5.so library.

    This path is used to check if the Kerberos library supports the optional
    APIs when compiling.

    Args:
        libraries: The libraries specified by krb5-config (-l).
        librariy_dirs: The library dirs specified by krb5-config (-L).
        link_args: The linking args specified by krb5-config.
        macos_native: Whether this is for the native macOS Kerberos library.

    Returns:
        str: The path to the krb5 library.
    """
    krb5_lib = os.environ.get("KRB5_MAIN_LIB", "")
    krb5_path = ""

    if not krb5_lib and macos_native:
        # Cannot use find_library as it won't look into the private framework path.
        krb5_lib = "/System/Library/PrivateFrameworks/Heimdal.framework/Heimdal"

    elif not krb5_lib:
        for opt in libraries:
            if opt.startswith("krb5"):
                ext = {
                    "nt": "dll",
                    "darwin": "dylib",
                }.get(os.name, "so")

                krb5_lib = f"lib{opt}.{ext}"
                break

        for opt in link_args:
            # To support Heimdal on Debian, read the linker path.
            if opt.startswith("-Wl,/"):
                krb5_path = opt[4:] + "/"
                break

        if not krb5_path:
            for d in library_dirs:
                if os.path.exists(os.path.join(d, krb5_lib)):
                    krb5_path = d
                    break

    if not krb5_lib:
        raise Exception(
            "Could not find the main krb5 shared library. Set this manually with the env var KRB5_MAIN_LIB."
        )

    return os.path.join(krb5_path, krb5_lib)


raw_extensions = []
if not SKIP_EXTENSIONS:
    kc = get_krb5_config()
    print(f"Using krb5-config at '{kc}'")

    macos_native = False
    if sys.platform == "darwin":
        mac_ver = [int(v) for v in platform.mac_ver()[0].split(".")]
        macos_native = mac_ver >= [10, 7, 0]

    compile_args, raw_link_args = [
        shlex.split(os.environ[e], posix=True) if e in os.environ else None
        for e in ["KRB5_COMPILER_ARGS", "KRB5_LINKER_ARGS"]
    ]
    if compile_args is None:
        if macos_native:
            compile_args = []

        else:
            compile_args = shlex.split(run_command(f"{kc} --cflags krb5"))

        if CYTHON_LINETRACE:
            compile_args.append("-DCYTHON_TRACE_NOGIL=1")

    if raw_link_args is None:
        if macos_native:
            raw_link_args = ["-framework", "Heimdal", "-F", "/System/Library/PrivateFrameworks"]

        else:
            raw_link_args = shlex.split(run_command(f"{kc} --libs krb5"))

    library_dirs, include_dirs, libraries, link_args = [], [], [], []
    for arg in raw_link_args:
        if arg.startswith("-L"):
            library_dirs.append(arg[2:])
        elif arg.startswith("-l"):
            libraries.append(arg[2:])
        elif arg.startswith("-I"):
            include_dirs.append(arg[2:])
        else:
            link_args.append(arg)

    if macos_native:
        # Because this is doing a naughty thing and linking against a private framework there are no krb5.h header files
        # available. A hack is to just compile Heimdal itself and include that during compilation.
        heimdal_dir = os.environ.get("KRB5_MACOS_HEIMDAL_DIR", "")
        if not heimdal_dir:
            heimdal_dir = os.path.join(os.path.dirname(__file__), "build_helpers", "heimdal")

        include_dirs.append(os.path.join(os.path.abspath(heimdal_dir), "include"))

    krb5_path = get_krb5_lib_path(libraries, library_dirs, link_args, macos_native)
    print(f"Using {krb5_path} as Kerberos module for platform checks")
    krb5 = ctypes.CDLL(krb5_path)

    if hasattr(krb5, "krb5_xfree"):
        compile_args.append("-DHEIMDAL_XFREE")

    for e in [
        "ccache",
        ("ccache_mit", "krb5_cc_dup"),
        ("ccache_match", "krb5_cc_cache_match"),
        ("ccache_support_switch", "krb5_cc_support_switch"),
        "cccol",
        "context",
        ("context_mit", "krb5_init_secure_context"),
        "creds",
        "creds_opt",
        ("creds_opt_heimdal", "krb5_get_init_creds_opt_set_default_flags"),
        ("creds_opt_mit", "krb5_get_init_creds_opt_set_out_ccache"),
        ("creds_opt_set_in_ccache", "krb5_get_init_creds_opt_set_in_ccache"),
        ("creds_opt_set_pac_request", "krb5_get_init_creds_opt_set_pac_request"),
        "exceptions",
        "keyblock",
        "kt",
        ("kt_mit", "krb5_kt_dup"),
        ("kt_heimdal", "krb5_kt_get_full_name"),
        ("kt_have_content", "krb5_kt_have_content"),
        "principal",
        ("principal_heimdal", "krb5_principal_get_realm"),
        "string",
        ("string_mit", "krb5_enctype_to_name"),
    ]:
        name = e
        canary = None
        if isinstance(e, tuple):
            name = e[0]
            if len(e) > 1:
                canary = e[1]

        ext = make_extension(
            f"krb5._{name}",
            module=krb5,
            canary=canary,
            extra_link_args=link_args,
            extra_compile_args=compile_args,
            include_dirs=include_dirs,
            library_dirs=library_dirs,
            libraries=libraries,
        )
        if ext:
            raw_extensions.append(ext)

setup(
    ext_modules=cythonize(
        raw_extensions,
        language_level=3,
        compiler_directives={"linetrace": CYTHON_LINETRACE},
    ),
)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import enum
import typing

from libc.stdint cimport int32_t, uintptr_t

from krb5._exceptions import Krb5Error

from krb5._context cimport Context
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    # krb5_free_unparsed_name is deprecated in Heimdal - use krb5_xfree instead
    """
    void krb5_free_unparsed_name_generic(krb5_context context, char *val)
    {
    #if defined(HEIMDAL_XFREE)
        krb5_xfree(val);
    #else
        krb5_free_unparsed_name(context, val);
    #endif
    }
    """

    krb5_error_code krb5_copy_principal(
        krb5_context context,
        krb5_const_principal inprinc,
        krb5_principal *outprinc,
    ) nogil

    void krb5_free_principal(
        krb5_context context,
        krb5_principal val,
    ) nogil

    # See inline C code
    void krb5_free_unparsed_name_generic(
        krb5_context context,
        char *val,
    ) nogil

    krb5_error_code krb5_parse_name_flags(
        krb5_context context,
        const char *name,
        int flags,
        krb5_principal *principal_out,
    ) nogil

    krb5_error_code krb5_unparse_name_flags(
        krb5_context context,
        krb5_const_principal principal,
        int flags,
        char **name,
    ) nogil

    int32_t KRB5_PRINCIPAL_PARSE_NO_REALM
    int32_t KRB5_PRINCIPAL_PARSE_REQUIRE_REALM
    int32_t KRB5_PRINCIPAL_PARSE_ENTERPRISE
    int32_t KRB5_PRINCIPAL_PARSE_IGNORE_REALM

    int32_t KRB5_PRINCIPAL_UNPARSE_SHORT
    int32_t KRB5_PRINCIPAL_UNPARSE_NO_REALM
    int32_t KRB5_PRINCIPAL_UNPARSE_DISPLAY


class PrincipalParseFlags(enum.IntEnum):
    none = 0
    no_realm = KRB5_PRINCIPAL_PARSE_NO_REALM
    require_realm = KRB5_PRINCIPAL_PARSE_REQUIRE_REALM
    enterprise = KRB5_PRINCIPAL_PARSE_ENTERPRISE
    ignore_realm = KRB5_PRINCIPAL_PARSE_IGNORE_REALM


class PrincipalUnparseFlags(enum.IntEnum):
    none = 0
    short = KRB5_PRINCIPAL_UNPARSE_SHORT
    no_realm = KRB5_PRINCIPAL_UNPARSE_NO_REALM
    display = KRB5_PRINCIPAL_UNPARSE_DISPLAY


cdef class Principal:
    """Kerberos Principal object.

    This class represents a Kerberos principal.

    Args:
        context: Krb5 context.
    """
    # cdef Context ctx
    # cdef krb5_principal raw
    # cdef int needs_free
    # cdef int _parse_flags

    def __cinit__(Principal self, Context context, flags, int needs_free=1):
        self.ctx = context
        self.raw = NULL
        self.needs_free = needs_free
        self._parse_flags = flags

    def __copy__(Principal self):
        return copy_principal(self.ctx, self)

    def __dealloc__(Principal self):
        if self.raw and self.needs_free:
            krb5_free_principal(self.ctx.raw, self.raw)
            self.raw = NULL

    @property
    def addr(Principal self) -> typing.Optional[int]:
        if self.raw:
            return <uintptr_t>self.raw

    @property
    def name(Principal self) -> typing.Optional[bytes]:
        if self.raw:
            # Heimdal fails to unparse a no_realm principal if the no_realm unparse flags aren't used.
            flags = PrincipalUnparseFlags.none
            if (
                self._parse_flags & PrincipalParseFlags.no_realm or
                self._parse_flags & PrincipalParseFlags.ignore_realm
            ):
                flags = PrincipalUnparseFlags.no_realm

            return unparse_name_flags(self.ctx, self, flags=flags)

    def __repr__(Principal self) -> str:
        name = self.name
        return f"Principal({name.decode('utf-8') if name else 'NULL'})"

    def __str__(Principal self) -> str:
        name = self.name
        return name.decode('utf-8') if name else 'NULL'


def copy_principal(
    Context context not None,
    Principal principal not None,
) -> Principal:
    out = Principal(context, principal._parse_flags)
    cdef krb5_error_code err = 0

    err = krb5_copy_principal(context.raw, principal.raw, &out.raw)
    if err:
        raise Krb5Error(context, err)

    return out


def parse_name_flags(
    Context context not None,
    const unsigned char[:] name not None,
    int flags=PrincipalParseFlags.none,
) -> Principal:
    principal = Principal(context, flags)
    cdef krb5_error_code err = 0

    cdef const char *name_ptr = NULL
    if name is not None and len(name):
        name_ptr = <const char*>&name[0]
    else:
        raise ValueError("Principal must be set")

    err = krb5_parse_name_flags(context.raw, name_ptr, flags, &principal.raw)
    if err:
        raise Krb5Error(context, err)

    return principal


def unparse_name_flags(
    Context context not None,
    Principal principal not None,
    int flags=PrincipalUnparseFlags.none,
) -> bytes:
    cdef krb5_error_code err = 0
    cdef char *name = NULL

    err = krb5_unparse_name_flags(context.raw, principal.raw, flags, &name)
    if err:
        raise Krb5Error(context, err)

    try:
        return <bytes>name
    finally:
        krb5_free_unparsed_name_generic(context.raw, name)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._krb5_types cimport *


cdef class Context:
    cdef krb5_context raw
